/**
 * @file keyword_hash.c
 * \brief Helper to hash keywords.
 *
 * Helper that hashes keywords and help to find them back.
 */

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include "hash.h"

/* Our link between GPMI and Hash module */

/**
 * \brief Hasher for string->(int *) pairs for static tables.
 */
unsigned int gpmi_hash_int_static(int cmd, void *data1, void *data2, char **result_pointer)
{
	switch (cmd) {
		case h2_cmd_freekey:
		case h2_cmd_allockey:
		case h2_cmd_key:
		case h2_cmd_cmpkey:
			return hash_defcalc(cmd, (char *)data1, (char *)data2, result_pointer);
		case h2_cmd_cmprec:
			if (*(int *)data1 == *(int *)data2)
				return 0;
			return 1;
	}
	return -1;
}

/**
 * \brief Hasher for string->(int *) pairs for dynamic tables
 */
unsigned int gpmi_hash_int_dynamic(int cmd, void *data1, void *data2, char **result_pointer)
{
	switch (cmd) {
		case h2_cmd_freekey:
		case h2_cmd_allockey:
		case h2_cmd_key:
		case h2_cmd_cmpkey:
			return hash_defcalc(cmd, (char *)data1, (char *)data2, result_pointer);
		case h2_cmd_cmprec:
			if (*(int *)data1 == *(int *)data2)
				return 0;
			return 1;
		case h2_cmd_destroy:
			free(((hdata_t *)data1)->data);
	}
	return -1;
}

/**
 * \brief Hasher for string->string pairs for dynamic tables.
 */
unsigned int gpmi_hash_str_dynamic(int cmd, void *data1, void *data2, char **result_pointer)
{
	switch (cmd) {
		case h2_cmd_freekey:
		case h2_cmd_allockey:
		case h2_cmd_key:
		case h2_cmd_cmpkey:
		case h2_cmd_cmprec:
			return hash_defcalc(cmd, (char *)data1, (char *)data2, result_pointer);
		case h2_cmd_destroy:
			free(((hdata_t *)data1)->data);
	}
	return -1;
}

/**
 * \brief Hasher for string->(void *) pairs.
 */
unsigned int gpmi_hash_ptr(int cmd, void *data1, void *data2, char **result_pointer)
{
	switch (cmd) {
		case h2_cmd_freekey:
		case h2_cmd_allockey:
		case h2_cmd_key:
		case h2_cmd_cmpkey:
			return hash_defcalc(cmd, (char *)data1, (char *)data2, result_pointer);
		case h2_cmd_cmprec:
			if (data1 == data2)
				return 0;
			return 1;
	}
	return -1;
}


/**
 * \brief Hasher for (int *)->string pairs.
 */
unsigned int gpmi_hash_int_rev(int cmd, void *data1, void *data2, char **result_pointer)
{
	int *i;
	switch (cmd) {
		case h2_cmd_key:
			return *(int *)data1;
		case h2_cmd_cmpkey:
			if (*(int *)data1 == *(int *)data2)
				return 0;
			return 1;
		case h2_cmd_cmprec:
			assert(data1 != NULL && data2 != NULL);
			if (strcasecmp((char *)data1, (char *)data2) == 0)
				return 0;
			return 1;
		case h2_cmd_allockey:
			i = malloc(sizeof(int));
			*i = *(int *)data1;
			*result_pointer = (char *)i;
			return 0;
		case h2_cmd_freekey:
			assert(data1 != NULL);
			free(data1);
			return 0;
	}
	return -1;
}

/**
 * \brief Hasher for (void *)->string pairs.
 */
unsigned int gpmi_hash_ptr_rev(int cmd, void *data1, void *data2, char **result_pointer)
{
	switch (cmd) {
		case h2_cmd_key:
			return (unsigned int)(size_t)data1;
		case h2_cmd_cmpkey:
			if (data1 == data2)
				return 0;
			return 1;
		case h2_cmd_cmprec:
			assert(data1 != NULL && data2 != NULL);
			if (strcasecmp((char *)data1, (char *)data2) == 0)
				return 0;
			return 1;
		case h2_cmd_allockey:
			*result_pointer = data1;
			return 0;
		case h2_cmd_freekey:
			return 0;
	}
	return -1;
}
