#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "libs.h"
#include "os_dep.h"
#include "error.h"

#ifdef WIN32
#include <windows.h>
#include <shlobj.h>
#endif

#if defined(WIN32) && !defined(__CYGWIN__)

/* Function basename is not native on windows */
char *basename(char *path)
{
	char* last = strrchr(path, '/');
	if ((path == NULL) || (*path == '\0')) return ".";
	if (*path == '/' && *(path + 1) == '\0') return "/";
	if (last == NULL) return path;
	if (last && *(last + 1) == '\0') {
		*last = '\0';
		last = strrchr(path, '/');
	}
	if (last) return (++last);
	return path;
}

/* Function dirname is not native on windows */
char *dirname(char *path)
{
	char* last = strrchr(path, '/');
	if ((path == NULL) || (*path == '\0')) return ".";
	if (*path == '/' && *(path + 1) == '\0') return "/";
	if (last == NULL) return ".";
	if (last && *(last + 1) == '\0') {
		*last = '\0';
		last = strrchr(path, '/');
	}
	if (last) *last = '\0';
	if (*path == '\0') return "/";
	return path;
}

#endif

/**
 * Build full pathname of a library and its dep
 */
void gpmi_library_add_path(const char *dir, const char *library_name, char *lib_file_name, char *dep_file_name, int path_size)
{

/* Os dependent part: the file name of a dynamic lib */
#ifdef DLCOMPAT
		/* With dlcompat, just add dir, file name and extension */
		snprintf(lib_file_name,	path_size,	"%s/%s.dlc",	dir, library_name);
		snprintf(dep_file_name,	path_size,	"%s/%s.dep",	dir, library_name);
#elif DARWIN
		/* On darwin, it is like unix, only so -> dylib */
		snprintf(lib_file_name, path_size,      "%s/%s.dylib",     dir, library_name);
		snprintf(dep_file_name, path_size,      "%s/%s.dep",    dir, library_name);
#else
		/* On unix, just add dir, file name and extension */
		snprintf(lib_file_name,	path_size,	"%s/%s.so",		dir, library_name);
		snprintf(dep_file_name,	path_size,	"%s/%s.dep",	dir, library_name);
#endif
}

/**
 *  Os dependent part for loading a dynamic linked library
 */
int gpmi_library_load_lowlevel(gpmi_library *library, char *file_name)
{
	char *error_str;
	library->handle = dlopen(file_name, RTLD_GLOBAL | RTLD_LAZY);
	if (library->handle == NULL) {
		error_str = malloc(256);
		snprintf(error_str, 256, "%s on file %s", dlerror(), file_name);
		gpmi_err_stack_push(gpmi_err_load_library, gpmi_err_src_lib_load_lowlevel, error_str);
		free(error_str);
		return 1;
	}

	return 0;
}

/**
 *  OS dependent part to unload a dynamic link library
 */
void gpmi_library_unload_lowlevel(gpmi_library *library)
{
	dlclose(library->handle);
}

/**
 *  OS dependent part to resolve a symbol
 */
void *gpmi_dlsym_lowlevel(gpmi_library *library, const char *name)
{
	return dlsym(library->handle, name);
}

/**
 *  OS dependent part to get a handle to the global address space
 */
void *gpmi_dlopen_global()
{
	return dlopen(NULL, RTLD_GLOBAL);
}

int gpmi_install_root(char *inst_root, int inst_root_size)
{
#ifdef WIN32
	HKEY hKey;
	LONG lRet;
	char buf[PATH_MAX];
	DWORD size;

	// check HKCU\SOFTWARE\LibGPMI\InstallPath
	lRet = RegOpenKeyEx(HKEY_CURRENT_USER, "SOFTWARE\\LibGPMI", 0, KEY_QUERY_VALUE, &hKey);
	if (lRet == ERROR_SUCCESS) {
		size = inst_root_size;
		lRet = RegQueryValueEx(hKey, "InstallPath", NULL, NULL, (LPBYTE)buf, &size);
		RegCloseKey(hKey);
		if (lRet == ERROR_SUCCESS) {
			snprintf(inst_root, inst_root_size, buf);
			return 0;
		}
	}

	// check HKLM\SOFTWARE\LibGPMI\InstallPath
	lRet = RegOpenKeyEx(HKEY_LOCAL_MACHINE, "SOFTWARE\\LibGPMI", 0, KEY_QUERY_VALUE, &hKey);
	if (lRet == ERROR_SUCCESS) {
		size = inst_root_size;
		lRet = RegQueryValueEx(hKey, "InstallPath", NULL, NULL, (LPBYTE)buf, &size);
		RegCloseKey(hKey);
		if (lRet == ERROR_SUCCESS) {
			snprintf(inst_root, inst_root_size, buf);
			return 0;
		}
	}
#endif
	snprintf(inst_root, inst_root_size, GPMI_INST_ROOT);
	return 0;
}

#ifdef WIN32
char *gpmi_init_path(char *path)
{
	char inst_root[PATH_MAX];
	char buf[PATH_MAX];

	gpmi_install_root(inst_root, sizeof(inst_root));
	snprintf(buf, sizeof(buf), "%s%s", inst_root, path);
	return strdup(buf);
}
#endif

char *gpmi_getenv(gpmi_env_t env)
{

#ifdef UNIX
	switch(env) {
		case gpmi_env_home:
			return strdup(getenv("HOME"));
	}
#endif
#ifdef WIN32
	switch(env) {
		case gpmi_env_home: {
			char path[MAX_PATH];
			SHGetFolderPath(NULL, CSIDL_APPDATA|CSIDL_FLAG_CREATE, NULL, 0, path);
			return strdup(path);
		}
	}
#endif
	assert(!"Invalid env in gpmi_getenv or porting problem");
	return NULL;
}
