#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#ifndef WIN32
#include <sys/sendfile.h>
#include <sys/types.h>
#include <sys/stat.h>
#else
#define fsync FlushFileBuffers
#endif
#include <assert.h>
#include "cli.h"
#include "filter.h"


#define delay 1
/* Timeout (unit is delay's value) */
#define read_timeout 100

#define GPMI_FIN     "###GPMI_FIN###"
#define GPMI_FIN_LEN 14
#define GPMI_CMD     "###GPMI### "
#define GPMI_CMD_LEN 11

/* Open a config file, find the language needed and parse its line. Returns 0 on success */
int cli_read_conf(const char *dir, const char *language, char **before, char **after)
{
	FILE *f;
	char *fn;
	char buff[256], *s, *s1, *s2;
	int sl;

	fn = malloc(strlen(dir) + 64);
	sprintf(fn, "%s/cli.conf", dir);

	f = fopen(fn, "r");
	free(fn);

	if (f == NULL)
		return -1;

	sl = strlen(language);

	while(!(feof(f))) {
		*buff = '\0';
		fgets(buff, sizeof(buff), f);
		switch (*buff) {
			case '\r':
			case '\n':
			case '#':
				/* Ignore comments and empty lines */
				break;
			default:
				s = buff;

				/* strip leading white space */
				while(isspace(*s))
					s++;

				/* the first char after the language name must be space, name must match */
				if (isspace(s[sl]) && (strncmp(s, language, sl) == 0)) {
					s += sl;
					/* strip leading white space */
					while(isspace(*s))
						s++;
					s1 = s;

					/* Find first non-space */
					while(!isspace(*s))
						s++;
					*s = '\0'; s++;

					/* Skip space between the two columns */
					while(isspace(*s))
						s++;
					s2 = s;

					/* Find first non-space */
					while(!isspace(*s))
						s++;
					*s = '\0';

					*before = strdup(s1);
					*after  = strdup(s2);
					fclose(f);
					return 0;
				}
		}
	}

	fclose(f);
	return -2;
}

/* Search for language in all places we may have a config in */
int cli_read_confs(const char *mod_path, const char *language, char **dir, char **before, char **after)
{
	int ret;

	*dir = malloc(strlen(mod_path) + 16);
	sprintf(*dir, "%s/cli-data", mod_path);
	ret = cli_read_conf(*dir, language, before, after);
	if (ret < 0) {
		free(*dir);
		*dir = NULL;
	}
	return ret;
}

int cli_read_confs_multi(int gpmi_path, const char *language, char **dir, char **before, char **after)
{
	int pos, ret;
	char *path;

	/* Check all module paths */
	for(pos = 0;;) {
		path = gpmi_path_get(gpmi_path, &pos);
		if (path == NULL)
			return -1;
		ret = cli_read_confs(path, language, dir, before, after);
		free(path);
		if (ret == 0)
			return 0;
	}

	/* We never get here but -Wall still would warn */
	return -2;
}


int cli_safe_tmp(FILE **f, char **fn)
{

	/* XXX: we need a read temp file name here */
	*fn = malloc(128);
	sprintf(*fn, "./gpmi.script.tmp");
	*f = fopen(*fn, "w");

	if (*f == NULL) {
		free(*fn);
		*fn = NULL;
		return -1;
	}

	return 0;
}

int cli_copy_file(FILE *from, FILE *to)
{
#ifndef WIN32
	off_t o;
#endif
	char buff[BUFSIZ];
	int from_, to_, len;

	from_ = fileno(from);
	to_   = fileno(to);

#ifndef WIN32
	/* This is the fast way: copy inside the kernel */
	o = 0;
	if (sendfile(to_, from_, &o, 0xFFFFFFFF) >= 0)
		return 0;
#endif

	/* Fallback: much slower but will run anywhere */
	for(;;) {
		len = read(from_, buff, sizeof(buff));
		if (len <= 0)
			return len;

		write(to_, buff, len);
	}

	/* We never get here, this return will supress a warning with -Wall */
	return -1;
}

int cli_append(FILE *f, const char *dir, const char *fn)
{
	char *fn_;
	FILE *from;

	if (dir != NULL) {
		fn_ = malloc(strlen(fn) + strlen(dir) + 4);
		sprintf(fn_, "%s/%s", dir, fn);
		from = fopen(fn_, "r");
		free(fn_);
	}
	else
		from = fopen(fn, "r");

	if (from == NULL)
		return -1;

	if (cli_copy_file(from, f) != 0) {
		fclose(from);
		return -2;
	}

	fclose(from);
	return 0;
}

cli_t *cli_fork(const char *dir, const char *before, const char *script, const char *after)
{
	FILE *f;
	char *fn;
	char *commandline;
	cli_t *c;

	if (cli_safe_tmp(&f, &fn) != 0)
		return NULL;

	/* Assamble the file */
	if (cli_append(f, dir, 	before) != 0)
		goto error;
	write(fileno(f), "\n\n", 2);
	if (cli_append(f, NULL,	script) != 0)
		goto error;
	write(fileno(f), "\n\n", 2);
	if (cli_append(f, dir, 	after) != 0)
		goto error;
	write(fileno(f), "\n\n", 2);

	/* Close the file so we can execute it */
	fclose(f);

	c = malloc(sizeof(cli_t));
	c->fn = strdup(fn);

	if (bgproc_create_pipe(c->fds_a, BGPROC_PIPE_WRITE) != 0)
		goto error_c;

	if (bgproc_create_pipe(c->fds_b, BGPROC_PIPE_READ) != 0)
		goto error_c;

#ifdef WIN32
	commandline = malloc(strlen(fn) + 4);
	sprintf(commandline, "sh %s", fn);
#else
	chmod(fn, S_IXUSR | S_IRUSR | S_IWUSR);
	commandline = fn;
#endif

	if (bgproc_create_child(commandline, c->fds_a, c->fds_b) != 0)
		printf("child creation failed\n");

	/* Set up the filter to read data from the script */
	c->flt_read  = flt_init("\r\n", 0, flt_oversep);
	c->flt_write = flt_init("\r\n", 0, flt_oversep);

	/* Start with idle state */
	c->state = cli_st_idle;

	/* Return on success */
	return c;

	/* Clean up and return NULL on error */
	error_c:;
		free(c->fn);
		free(c);
	error:;
		fclose(f);
		free(fn);
		return NULL;
}

int cli_parse(cli_t *c, int timeout)
{
	int len;
	char *s, buff[1024];
	char *wrt;

	/* If we are not waiting for the last operation to finish, push the next message and wait */
	if (c->state != cli_st_waitfin) {
		cli_dprintf("[cli] No need to wait fin\n");
		len = flt_fill(c->flt_write, NULL, 0, &wrt);
		if (len <= 0) {
//			fprintf(stderr, "[cli] Nothing to write, returning\n");
//			return 0;
		}
		else {
			cli_dprintf("[cli]   pushing1: '%s'\n", wrt);
/*			bgproc_write_to_pipe(c->fds_a[1], wrt, len);
			bgproc_write_to_pipe(c->fds_a[1], "\n", 1);*/
			cli_send_string_now(c, wrt);
			cli_send_string_now(c, "\n");
			c->state = cli_st_waitfin;
		}
	}
	else
		cli_dprintf( "[cli] waiting for the previous fin\n");

	cli_dprintf( "[cli] cycle\n");

	while(timeout > 0) {
		len = bgproc_read_from_pipe(c->fds_b[0], buff, sizeof(buff));
		len = flt_fill(c->flt_read, buff, len, &s);
		if (s != NULL) {
			cli_dprintf( "[cli] Read: '%s'\n", s);
			if (strncmp(s, GPMI_FIN, GPMI_FIN_LEN) == 0) {
				cli_dprintf( "[cli] read fin.\n");
				len = flt_fill(c->flt_write, NULL, 0, &wrt);
				if (len <= 0) {
					cli_dprintf( "[cli]   nothing left in the write fifo.\n");
					c->state = cli_st_idle;
					return 0;
				}
				cli_dprintf( "[cli]   pushing2: '%s'\n", wrt);
/*				bgproc_write_to_pipe(c->fds_a[1], wrt, len);
				bgproc_write_to_pipe(c->fds_a[1], '\n', 1);*/
				cli_send_string_now(c, wrt);
				cli_send_string_now(c, "\n");
				c->state = cli_st_waitfin;
			}
			else if (strncmp(s, GPMI_CMD, GPMI_CMD_LEN) == 0) {
				cli_dprintf( "[cli] dynamic command '%s'\n", s+11);
				Cli_DynamicPackageCommand(s + 11);
			}
			else
				cli_dprintf( "[cli] read junk: '%s'\n", s);
		}
		#ifdef WIN32
			Sleep(delay / 1000);
		#else
			usleep(delay);
		#endif
		timeout--;
	}
	cli_dprintf( "[cli] timeout.\n");
	return -1;
}

void cli_send_string(cli_t *c, const char *s)
{
	int n;

	/* We don't really write here, only push the string into the fifo */
	n = flt_fill(c->flt_write, (char *)s, strlen(s), NULL);
	cli_dprintf( "[cli] putting in fifo: '%s'\n", s);
}

void cli_send_string_now(cli_t *c, const char *s)
{
	cli_dprintf( "[cli] writing: '%s'\n", s);
	bgproc_write_to_pipe(c->fds_a[1], s, strlen(s));
	fsync(c->fds_a[1]);
}
