#include "perl.h"

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <XSUB.h>

#include "gpmi/debug.h"

/**
 * Returns the extension of this script-module.
 */
char *Perl_GetExtension()
{
	return strdup(".pl");
}

/**
 * Reloads the script on the fly.
 */
int Perl_Rehash(void *client_data)
{
	/* TODO */
	return 0;
}

extern void boot_DynaLoader (pTHX_ CV* cv);

static void xs_init(pTHX)
{
	char *file = __FILE__;
	dXSUB_SYS;

	/* DynaLoader is a special case */
	newXS("DynaLoader::boot_DynaLoader", boot_DynaLoader, file);
}

/**
 * Execute a dynamic command.
 **/
XS(Perl_DynamicPackageCommand)
{
	//Perl_ClientData *Perl_MyData = GPMI_Script_GetClientData();
	char *(*proc)(int argc, char *argv[]);
	GV *gv;
	SV* sub_name;
	char *function_name, *tmp_name, *s;
	char **args;
	dXSARGS;
	int i;

	SP -= items;

	args = malloc(sizeof(char *) * (items + 1));
	for (i = 0; i < items; i++)
		args[i + 1] = SvPV_nolen(ST(i));

	/* We can not find the current functionname the normal way, so I just copied this from
	 *  the source-code, where the functionname lookup is done... easy as pie ;) */
	gv = CvGV(cv);
	sub_name = sv_newmortal();
	gv_efullname3(sub_name, gv, Nullch);
	tmp_name = SvPV_nolen(sub_name);
	/* Copy the function-name, but skip the first 5 chars (always: main:) */
	function_name = malloc(strlen(tmp_name) - 6 + 1);
	for (i = 6; i < strlen(tmp_name) + 1; i++)
		function_name[i - 6] = tmp_name[i];
	args[0] = function_name;

	/* Find the proc */
	proc = GPMI_Script_FindProc(function_name);

	/* Check if command not found - this must be an internal error: the
	   package had registered the command in the hash, but the routine
	   pointer is NULL.
	*/
	if (proc == NULL) {
		gpmi_debug(0, "[Perl] Internal error: %s has a NULL proc in the command hash\n", function_name);
		return XSRETURN_EMPTY;
	}

	/* Execute the command */
	s = proc(items + 1, args);

	free(function_name);
	free(args);

	/* Return the right value */
	if (s == NULL) {
		return XSRETURN_EMPTY;
	}

	ST(0) = newSVpv(s, 0);
	sv_2mortal(ST(0));
	free(s);
	return XSRETURN(1);
}

/**
 * Loads a script.
 */
int Perl_LoadScript(void *client_data, const char *filename)
{
	Perl_ClientData *Perl_MyData = client_data;
	char *embedding[2];
	int i;

	embedding[0] = NULL;
	embedding[1] = strdup(filename);

	/* Load the script */
	PERL_SET_CONTEXT(Perl_MyData->interp);
	perl_parse(Perl_MyData->interp, xs_init, 2, embedding, NULL);

	/* Dequeue any queued commands (to register) */
	for (i = 0; i < Perl_MyData->commands_size; i++)
		newXS(Perl_MyData->commands[i], Perl_DynamicPackageCommand, __FILE__);

	/* Mark that we are running, this way we don't queue commands any longer */
	Perl_MyData->started = 1;

	/* Run the script */
	perl_run(Perl_MyData->interp);

	free(embedding[1]);

	return 0;
}

/**
 * Register a command in our command hash
 */
void Perl_RegisterCommand(void *client_data, const char *name)
{
	Perl_ClientData *Perl_MyData = client_data;

	if (Perl_MyData->started != 1) {
		/* We can not add commands to Perl if we haven't a script loaded.. so save the names
		 *  temporary in an array, and add them after the script is loaded */
		Perl_MyData->commands_size++;
		Perl_MyData->commands = realloc(Perl_MyData->commands, sizeof(char *) * Perl_MyData->commands_size);

		Perl_MyData->commands[Perl_MyData->commands_size - 1] = strdup(name);
		return;
	}

	/* Register the command */
	PERL_SET_CONTEXT(Perl_MyData->interp);
	newXS((char *)name, Perl_DynamicPackageCommand, __FILE__);
}

/**
 * Call a function of the script by name, passing params.
 */
void Perl_RunFunction(void *client_data, const char *function, int argc, const char **argv)
{
	Perl_ClientData *Perl_MyData = client_data;
	char **args;
	int i;

	/* Temporary copy the string, because we need a null-terminated array */
	args = malloc((argc + 1) * sizeof(char *));

	for (i = 0; i < argc; i++)
		args[i] = strdup(argv[i]);
	args[argc] = NULL;

	/* Call the Perl function */
	PERL_SET_CONTEXT(Perl_MyData->interp);
	call_argv(function, G_DISCARD, args);

	/* Free the temporary stuff */
	for (i = 0; i < argc; i++)
		free(args[i]);
	free(args);
}

int Perl_Initialize(void **client_data)
{
	Perl_ClientData *MyData;

	MyData = malloc(sizeof(Perl_ClientData));
	memset(MyData, 0, sizeof(Perl_ClientData));

	/* Start up a perl interpreter */
	MyData->interp = perl_alloc();
	if (MyData->interp != NULL) {
		PERL_SET_CONTEXT(MyData->interp);
		perl_construct(MyData->interp);
	}

	/* Stop if that failed */
	if (MyData->interp == NULL) {
		free(MyData);
		return -1;
	}

	*client_data = MyData;

	return 0;
}

void Perl_Destroy(void **client_data)
{
	Perl_ClientData *Perl_MyData = *client_data;

	if (Perl_MyData->interp != NULL) {
		/* Kill the perl instance */
		PERL_SET_CONTEXT(Perl_MyData->interp);
		perl_destruct(Perl_MyData->interp);
		perl_free(Perl_MyData->interp);
	}

	if (Perl_MyData->commands != NULL) {
		/* Free the command-queue */
		int i;
		for (i = 0; i < Perl_MyData->commands_size; i++)
			free(Perl_MyData->commands[i]);
		free(Perl_MyData->commands);
	}

	/* Finally free up the client-data */
	free(*client_data);
}
